// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "actgen_dribble_table_omnidir_dribble.h"
#include <rcsc/action/dribble_table2009.h>

#include <rcsc/common/logger.h>

void
ActGen_DribbleTableOmnidirDribble::addCandidates( std::vector< rcsc::ActionStatePair > * result,
                                                  const rcsc::PredictState & state,
                                                  const rcsc::WorldModel & current_wm,
                                                  const std::vector< rcsc::ActionStatePair > & path ) const
{
    if ( ! path.empty() )
    {
        return;
    }


    //
    // check game mode
    //
    if ( state.gameMode().type() != rcsc::GameMode::PlayOn )
    {
        return;
    }


    //
    // get dribble table
    //
    rcsc::DribbleTable2009 & table = rcsc::DribbleTable2009::instance();
    table.update( current_wm );

    const std::vector< rcsc::CooperativeAction::Ptr > &
        dribbleCandidates = table.getDribbleCandidates( current_wm );


    //
    // add dribble candidates
    //
    const std::vector< rcsc::CooperativeAction::Ptr >::const_iterator end = dribbleCandidates.end();
    for ( std::vector< rcsc::CooperativeAction::Ptr >::const_iterator act = dribbleCandidates.begin();
          act != end;
          ++act )
    {
        result->push_back( rcsc::ActionStatePair
                           ( *act,
                             new rcsc::PredictState
                             ( state,
                               (*act)->ballReachStep(),
                               (*act)->targetPlayerUnum(),
                               (*act)->targetPoint() ) ) );
    }
}
