// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "actgen_pass_table_pass.h"
#include "pass_table2009.h"
#include "simple_pass_checker.h"
#include <rcsc/common/logger.h>

//#define DEBUG_PRINT

void
ActGen_PassTablePass::addCandidates( std::vector< rcsc::ActionStatePair > * result,
                                     const rcsc::PredictState & state,
                                     const rcsc::WorldModel & current_wm,
                                     const std::vector< rcsc::ActionStatePair > & path ) const
{
    if ( ! path.empty() )
    {
        return;
    }

//     const rcsc::SimplePassChecker pass_check;

    const std::vector< rcsc::CooperativeAction::Ptr > &
        passCandidates = rcsc::PassTable2009::instance().getPassCandidates( current_wm );

    //
    // add pass candidates
    //
    const std::vector< rcsc::CooperativeAction::Ptr >::const_iterator end = passCandidates.end();
    for ( std::vector< rcsc::CooperativeAction::Ptr >::const_iterator act = passCandidates.begin();
          act != end;
          ++act )
    {
        if ( (*act)->targetPlayerUnum() == rcsc::Unum_Unknown )
        {
#ifdef DEBUG_PRINT
            rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                                __FILE__": can't pass from %d,"
                                " target invalid",
                                state.ballHolder()->unum() );
#endif
            continue;
        }

        const rcsc::AbstractPlayerObject * target_player = state.teammate( (*act)->targetPlayerUnum() );

//         if ( ! pass_check( state, holder, target_player, (*act)->targetPoint(), (*act)->ballSpeed() ) )
//         {
// #ifdef DEBUG_PRINT
//             rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
//                                 "pass table: can't pass from %d to %d",
//                                 state.ballHolder()->unum(), target_player->unum() );
// #endif
//             continue;
//         }

        result->push_back( rcsc::ActionStatePair
                           ( *act,
                             new rcsc::PredictState
                             ( state,
                               (*act)->ballReachStep(),
                               target_player->unum(),
                               (*act)->targetPoint() ) ) );
    }
}
