// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef ACTGEN_RECURSIVE_LEVEL_FILTER_H
#define ACTGEN_RECURSIVE_LEVEL_FILTER_H

#include "cooperative_action_candidate_generator.h"
#include <boost/scoped_ptr.hpp>
#include <vector>

class ActGen_MaxRecursiveLevelFilter
    : public rcsc::CooperativeActionCandidateGenerator {
private:
    const boost::scoped_ptr< const rcsc::CooperativeActionCandidateGenerator > M_generator;
    const unsigned int M_max_threshold_level;

public:
    ActGen_MaxRecursiveLevelFilter( const rcsc::CooperativeActionCandidateGenerator * generator,
                                    unsigned int max_threshold_level )
        : M_generator( generator )
        , M_max_threshold_level( max_threshold_level )
      { }

    void addCandidates( std::vector< rcsc::ActionStatePair > * result,
                        const rcsc::PredictState & state,
                        const rcsc::WorldModel & current_wm,
                        const std::vector< rcsc::ActionStatePair > & path ) const
      {
          if ( path.size() < M_max_threshold_level )
          {
              M_generator->addCandidates( result, state, current_wm, path );
          }
      }
};

class ActGen_MinRecursiveLevelFilter
    : public rcsc::CooperativeActionCandidateGenerator {
private:
    const boost::scoped_ptr< const rcsc::CooperativeActionCandidateGenerator > M_generator;
    const unsigned int M_min_threshold_level;

public:
    ActGen_MinRecursiveLevelFilter( const rcsc::CooperativeActionCandidateGenerator * generator,
                                    unsigned int min_threshold_level )
        : M_generator( generator )
        , M_min_threshold_level( min_threshold_level )
      { }

    void addCandidates( std::vector< rcsc::ActionStatePair > * result,
                        const rcsc::PredictState & state,
                        const rcsc::WorldModel & current_wm,
                        const std::vector< rcsc::ActionStatePair > & path ) const
      {
          if ( path.size() + 1 >= M_min_threshold_level )
          {
              M_generator->addCandidates( result, state, current_wm, path );
          }
      }
};


class ActGen_RangeRecursiveLevelFilter
    : public rcsc::CooperativeActionCandidateGenerator {
public:
    static const int MIN = 1;
    static const int MAX = -1;

private:
    const boost::scoped_ptr< const rcsc::CooperativeActionCandidateGenerator > M_generator;

    const int M_min_threshold_level;
    const int M_max_threshold_level;

public:
    ActGen_RangeRecursiveLevelFilter( const rcsc::CooperativeActionCandidateGenerator * generator,
                                      unsigned int min_threshold_level,
                                      unsigned int max_threshold_level )
        : M_generator( generator )
        , M_min_threshold_level( min_threshold_level )
        , M_max_threshold_level( max_threshold_level )
      { }

    void addCandidates( std::vector< rcsc::ActionStatePair > * result,
                        const rcsc::PredictState & state,
                        const rcsc::WorldModel & current_wm,
                        const std::vector< rcsc::ActionStatePair > & path ) const
      {
          if ( ( M_max_threshold_level == MAX
                 || static_cast< int >( path.size() ) < M_max_threshold_level )
               && (static_cast< int >( path.size() + 1 ) >= M_min_threshold_level ) )
          {
              M_generator->addCandidates( result, state, current_wm, path );
          }
      }
};

#endif
