// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "actgen_shoot.h"
#include "default_field_evaluator.h"
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>

//#define DEBUG_PRINT

static const int VALID_PLAYER_THRESHOLD = 10;

ActGen_Shoot::ActGen_Shoot()
{
}

ActGen_Shoot::~ActGen_Shoot()
{
}


void
ActGen_Shoot::addCandidates( std::vector< rcsc::ActionStatePair > * result,
                             const rcsc::PredictState & state,
                             const rcsc::WorldModel &,
                             const std::vector< rcsc::ActionStatePair > & ) const
{
    const rcsc::AbstractPlayerObject * holder = state.ballHolder();

    if ( holder->posCount() > VALID_PLAYER_THRESHOLD
         || holder->isGhost() )
    {
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "shoot: unum %d can't shoot, pos accuracy low",
                            holder->unum() );
#endif
        return;
    }


    if ( ! DefaultFieldEvaluator::can_shoot_from( state, holder->pos(), VALID_PLAYER_THRESHOLD ) )
    {
        //
        // failure
        //
#ifdef DEBUG_PRINT
        rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                            "shoot: unum %d can't shoot",
                            holder->unum() );
#endif
        return;
    }


    //
    // success
    //
#ifdef DEBUG_PRINT
    rcsc::dlog.addText( rcsc::Logger::RECURSIVE,
                        "shoot: unum %d OK",
                        holder->unum() );
#endif

    const long shoot_spend_time
        = ( holder->pos().dist( rcsc::ServerParam::i().theirTeamGoalPos() ) / 1.5 );


    result->push_back( rcsc::ActionStatePair
                       ( new rcsc::CooperativeAction
                         ( rcsc::CooperativeAction::ActionType::Shoot,
                           rcsc::ServerParam::i().theirTeamGoalPos(),
                           holder->unum(),
                           shoot_spend_time,
                           shoot_spend_time,
                           true,
                           "shoot" ),
                         new rcsc::PredictState( state,
                                                 shoot_spend_time,
                                                 rcsc::ServerParam::i().theirTeamGoalPos() ) ) );
}
