// -*-c++-*-

/*!
  \file bhv_simple_dribble.h
  \brief simple dribble action
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_ACTION_BHV_SIMPLE_DRIBBLE_H
#define RCSC_ACTION_BHV_SIMPLE_DRIBBLE_H

#include <rcsc/player/cooperative_action.h>
#include <rcsc/player/soccer_action.h>
#include <rcsc/geom/vector_2d.h>

namespace rcsc {

/*!
  \class Bhv_SimpleDribble
  \brief simple dribble action
*/
class Bhv_SimpleDribble
    : public SoccerBehavior {
private:
    Vector2D M_target_point;
    int M_total_step;

    NeckAction * M_neck_action;
    ViewAction * M_view_action;

public:

    Bhv_SimpleDribble( const CooperativeAction & action,
                       NeckAction * neck = static_cast< NeckAction * >( 0 ),
                       ViewAction * view = static_cast< ViewAction * >( 0 ) );

    /*!
      \brief destruct subactions
     */
    ~Bhv_SimpleDribble();

    /*!
      \brief do dribble
      \param agent agent pointer to the agent itself
      \return true with action, false if can't do dribble
     */
    bool execute( PlayerAgent * agent );
};

}

#endif
