// -*-c++-*-

/*!
  \file pass_table.h
  \brief strict checking pass cource cache table Header File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA, Hiroki Shimora

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef HELIOS_PASS_TABLE_2009_H
#define HELIOS_PASS_TABLE_2009_H

#include <rcsc/player/world_model.h>
#include <rcsc/geom/vector_2d.h>
#include <rcsc/game_time.h>
#include <rcsc/player/cooperative_action.h>

#include <boost/shared_ptr.hpp>

#include <vector>
#include <set>

namespace rcsc {

/*!
  \class PassTable2009
  \brief strict checking pass cource cache.
 */
class PassTable2009 {
private:

    int M_total_count;
    std::vector< CooperativeAction::Ptr > M_courses;
    std::set< int > M_receive_point_set;


    // private for singleton
    PassTable2009();

    // not used
    PassTable2009( const PassTable2009 & );
    PassTable2009 & operator=( const PassTable2009 & );

public:

    static
    PassTable2009 & instance();

    void update( const WorldModel & wm );

    const
    std::vector< CooperativeAction::Ptr > & getPassCandidates( const WorldModel & wm );

private:

    void createDirectPass( const WorldModel & wm,
                           const Vector2D & first_ball_pos,
                           const AbstractPlayerCont & teammates );
    void createDirectLeadPass( const WorldModel & wm,
                               const Vector2D & first_ball_pos,
                               const AbstractPlayerObject * receiver,
                               const Vector2D & first_target_point );

    void createVoronoiPass( const WorldModel & wm,
                            const AbstractPlayerObject * ball_holder,
                            const Vector2D & first_ball_pos,
                            const AbstractPlayerCont & teammates );
    void createVoronoiLeadPass( const WorldModel & wm,
                                const AbstractPlayerObject * ball_holder,
                                const Vector2D & first_ball_pos,
                                const AbstractPlayerObject * receiver,
                                const Vector2D & first_target_point );

    void createSelfPass( const WorldModel & wm,
                         const AbstractPlayerCont & teammates );
    //void createSelfPass( const WorldModel & wm,
    //                     const AbstractPlayerObject * ball_holder );


    int predictReceiverReachStep( const WorldModel & wm,
                                  const AbstractPlayerObject * receiver,
                                  const Vector2D & receive_pos,
                                  const Vector2D & first_ball_pos,
                                  const AngleDeg & first_ball_vel_angle );

    bool checkCourse( const WorldModel & wm ,
                      const AbstractPlayerObject * receiver,
                      const Vector2D & receive_pos,
                      const double & ball_first_speed,
                      const int ball_reach_step );

    /*!
      \param target_angle == ball_first_vel.th()
     */
    bool existInterceptor( const PlayerObject & opponent,
                           const Vector2D & ball_first_pos,
                           const Vector2D & ball_first_vel,
                           const AngleDeg & target_angle,
                           const int ball_reach_step,
                           double * min_dist );
};

}

#endif
