// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa Akiyama, Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "pass_test_field_evaluator.h"
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/player/world_model.h>
#include <rcsc/geom/angle_deg.h>

#include <cmath>
#include <vector>
#include <algorithm>

using namespace rcsc;

//#define DEBUG_PRINT_SCORE_COLOR

/*-------------------------------------------------------------------*/
/*!

*/
PassTestFieldEvaluator::PassTestFieldEvaluator()
{
}

/*-------------------------------------------------------------------*/
/*!

*/
PassTestFieldEvaluator::~PassTestFieldEvaluator()
{
}


/*-------------------------------------------------------------------*/
/*!

*/
double
PassTestFieldEvaluator::operator()
                   ( const rcsc::PredictState & state,
                     const rcsc::WorldModel & world,
                     const std::vector< rcsc::ActionStatePair > & path ) const
{
#ifdef DEBUG_PRINT_SCORE_COLOR
    static const int color_tab[5][3] = {
        {   0,   0, 255 },
        {   0, 255, 255 },
        {   0, 255,   0 },
        { 255, 255,   0 },
        { 255,   0,   0 }
    };
#endif

    const rcsc::AbstractPlayerObject * receiver = state.ballHolder();

    const AngleDeg min_angle = -45.0;
    const AngleDeg max_angle = 45.0;
    const Vector2D goal_pos( 43.0, 0.0 );
#if 0
    const double ball_speed_max = ServerParam::i().ballSpeedMax();
    const double ball_speed_thr = std::min( 2.5, ball_speed_max - 0.1 );
#endif

    bool opp_very_near = false;
    const PlayerObject * nearest_opp = world.getOpponentNearestToSelf( 0 );
    if ( nearest_opp
         && nearest_opp->distFromSelf() < 2.0 )
    {
        opp_very_near = true;
    }

    double min_score = +100000.0;
    double max_score = -100000.0;

    int counter = 1;

    double score;

    {
        //-----------------------------------------------------------
        double opp_dist_rate = 1.0;
        {
            double opp_dist = 100.0;
            world.getOpponentNearestTo( receiver->pos(), 20, &opp_dist );
            //opp_dist_rate = std::pow( 0.99, std::max( 0.0, 30.0 - opp_dist ) );
            //opp_dist_rate = 1.0 - std::exp( - opp_dist*opp_dist / ( 2.0 * 20.0 ) );
            //opp_dist_rate = std::min( 1.0, std::sqrt( opp_dist * 0.1 ) );
            opp_dist_rate = std::min( 1.0, std::sqrt( opp_dist * 0.09 ) );
        }
        //-----------------------------------------------------------
        double goal_dist_rate = std::pow( 0.98,
                                          receiver->pos().dist( goal_pos ) );

        //-----------------------------------------------------------
        double x_diff_rate = 1.0;
        {
            double x_diff = receiver->pos().x - receiver->pos().x;
            if ( x_diff < 0.0 )
            {
                x_diff_rate = std::pow( 0.98, -x_diff );
            }
        }
        //-----------------------------------------------------------
        double pos_conf_rate = std::pow( 0.98, receiver->posCount() );
        //-----------------------------------------------------------
        double dir_conf_rate = 1.0;
        {
            AngleDeg pass_angle = ( receiver->pos() - world.self().pos() ).th();
            int max_dir_count = 0;
            world.dirRangeCount( pass_angle, 20.0, &max_dir_count, NULL, NULL );

            dir_conf_rate = std::pow( 0.98, max_dir_count );
        }
        //-----------------------------------------------------------
        //double offense_rate
        //    = std::pow( 0.98,
        //                std::max( 5.0, std::fabs( receiver->pos().y
        //                                          - world.ball().pos().y ) ) );
        //-----------------------------------------------------------
        score = 1000.0;
#if 0
        if ( opp_very_near && it->one_step_kick_ )
        {
            score += 1000.0;
        }
#endif
        score *= opp_dist_rate;
        score *= goal_dist_rate;
        score *= x_diff_rate;
        score *= pos_conf_rate;
        score *= dir_conf_rate;

#if 0
        if ( it->type_ == LEAD )
        {
            score *= 0.9;
        }

        if ( it->one_step_kick_ )
        {
            score *= 1.2;
        }
        else if ( it->first_speed_ > ball_speed_thr )
        {
            score *= 0.9;
        }
#endif

#if 1
        score -= path.size() * 1.0;
#endif

        if ( min_score > score ) min_score = score;
        if ( max_score < score ) max_score = score;

#if 0
        dlog.addText( Logger::PASS,
                      "<<<PASS %d Score %6.2f -- to%d(%.1f %.1f) recv_pos(%.1f %.1f) type %d "
                      " speed=%.2f",
                      counter,
                      score,
                      receiver->unum(),
                      receiver->pos().x, receiver->pos().y,
                      receiver->pos().x, receiver->pos().y,
                      it->type_,
                      it->first_speed_ );
        dlog.addText( Logger::PASS,
                      "____Rate: opp_dist=%.2f"
                      " goal_dist=%.2f"
                      " x_diff=%.2f"
                      " pos_conf=%.2f"
                      " dir_conf=%.2f %s",
                      opp_dist_rate,
                      goal_dist_rate,
                      x_diff_rate,
                      pos_conf_rate,
                      dir_conf_rate,
                      ( it->one_step_kick_ ? "one_step" : "" ) );
#else
        dlog.addText( Logger::PASS,
                      "<<<PASS %d Score %6.2f -- to%d(%.1f %.1f) recv_pos(%.1f %.1f)",
                      counter,
                      score,
                      receiver->unum(),
                      receiver->pos().x, receiver->pos().y,
                      receiver->pos().x, receiver->pos().y );
        dlog.addText( Logger::PASS,
                      "____Rate: opp_dist=%.2f"
                      " goal_dist=%.2f"
                      " x_diff=%.2f"
                      " pos_conf=%.2f"
                      " dir_conf=%.2f",
                      opp_dist_rate,
                      goal_dist_rate,
                      x_diff_rate,
                      pos_conf_rate,
                      dir_conf_rate );
#endif
    }

#ifdef DEBUG_PRINT_SCORE_COLOR
    min_score -= 10.0;
    max_score += 10.0;

    const double score_range = ( max_score == min_score
                                 ? 500.0
                                 : max_score - min_score );

    dlog.addText( Logger::PASS,
                  "__ score range = [%f, %f] %f",
                  min_score, max_score, score_range );

    {
        double score_rate = min_max( 0.0,
                                     ( score - min_score ) / score_range,
                                     0.99999999999 );
        int floor_index = static_cast< int >( std::floor( score_rate * 4.0 ) );
        double color_rate = min_max( 0.0, ( score_rate - 0.2 * floor_index ) / 0.2, 1.0 );
        int r = std::min( 255,
                          static_cast< int >( rint( color_tab[floor_index][0] * ( 1.0 - color_rate )
                                                    + color_tab[floor_index+1][0] * color_rate ) ) );
        int g = std::min( 255,
                          static_cast< int >( rint( color_tab[floor_index][1] * ( 1.0 - color_rate )
                                                    + color_tab[floor_index+1][1] * color_rate ) ) );
        int b = std::min( 255,
                          static_cast< int >( rint( color_tab[floor_index][2] * ( 1.0 - color_rate )
                                                    + color_tab[floor_index+1][2] * color_rate ) ) );
        dlog.addText( Logger::PASS,
                      "____ score_rate=%f floor_index=%d color_rate=%f color=(%d %d %d)",
                      score_rate, floor_index, color_rate,
                      r, g, b );
        dlog.addCircle( Logger::PASS,
                        receiver->pos(), 0.2,
                        r, g, b, true ); // filled circle
    }
#endif

    return score;
}
