// -*-c++-*-

/*!
  \file dribble_table2009.h
  \brief dribble course cache table Header File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 3 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef RCSC_ACTION_DRIBBLE_TABLE_2009_H
#define RCSC_ACTION_DRIBBLE_TABLE_2009_H

#include <rcsc/player/cooperative_action.h>
#include <rcsc/player/world_model.h>

#include <vector>

namespace rcsc {

class DribbleTable2009 {
private:
    int M_total_count;
    std::vector< CooperativeAction::Ptr > M_courses;
    std::vector< CooperativeAction::Ptr > M_aggressive_courses;

    // not used
    DribbleTable2009();
    DribbleTable2009( const DribbleTable2009 & );
    DribbleTable2009 & operator=( const DribbleTable2009 & );
public:

    static
    DribbleTable2009 & instance();

    void update( const WorldModel & wm );

    const
    std::vector< CooperativeAction::Ptr > & getDribbleCandidates( const WorldModel & wm );

    const
    std::vector< CooperativeAction::Ptr > & getAggressiveDribbleCandidates( const WorldModel & wm );

private:

    void createTargetPoints( const WorldModel & wm );
    void simulateKickTurnsDashes( const WorldModel & wm,
                                  const AngleDeg & dash_angle,
                                  const int n_turn );
    void createSelfCache( const WorldModel & wm,
                          const AngleDeg & dash_angle,
                          const int n_turn,
                          const int n_dash,
                          std::vector< Vector2D > & self_cache );
    bool checkOpponent( const WorldModel & wm,
                        const Vector2D & ball_trap_pos,
                        const int dribble_step,
                        const bool aggressive_mode );
};

}

#endif
