// -*-c++-*-

/*!
  \file field_analyzer.h
  \brief miscellaneous field analysis class Header File
*/

/*
 *Copyright:

  Copyright (C) Hidehisa AKIYAMA

  This code is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3, or (at your option)
  any later version.

  This code is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this code; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef HELIOS_FIELD_ANALYZER_H
#define HELIOS_FIELD_ANALYZER_H

#include "player_graph.h"

#include <rcsc/geom/voronoi_diagram.h>
#include <rcsc/geom/vector_2d.h>

class FieldAnalyzer {
private:

    rcsc::VoronoiDiagram M_all_players_voronoi_diagram;
    rcsc::VoronoiDiagram M_teammates_voronoi_diagram;
    rcsc::VoronoiDiagram M_pass_voronoi_diagram;
    PlayerGraph M_field_player_graph;

    FieldAnalyzer();
public:

    static
    FieldAnalyzer & instance();

    static
    const
    FieldAnalyzer & i()
      {
          return instance();
      }

    static
    int predict_self_reach_cycle( const rcsc::WorldModel & wm,
                                  const rcsc::Vector2D & target_point,
                                  const double & dist_thr,
                                  const int wait_cycle,
                                  const bool save_recovery,
                                  double * stamina );

    static
    int predit_player_turn_cycle( const rcsc::WorldModel & wm,
                                  const rcsc::AbstractPlayerObject * player,
                                  const rcsc::Vector2D & target_point,
                                  const int total_cycle,
                                  const double & dist_thr );

    static
    int predit_player_turn_cycle( const rcsc::WorldModel & wm,
                                  const rcsc::AbstractPlayerObject * player,
                                  const rcsc::AngleDeg & target_angle,
                                  const double & target_dist,
                                  const double & dist_thr );

    static
    int predict_player_reach_cycle( const rcsc::WorldModel & wm,
                                    const rcsc::AbstractPlayerObject * player,
                                    const rcsc::Vector2D & target_point,
                                    const double & dist_thr,
                                    const int wait_cycle );

    const
    rcsc::VoronoiDiagram & allPlayersVoronoiDiagram() const
      {
          return M_all_players_voronoi_diagram;
      }

    const
    rcsc::VoronoiDiagram & teammatesVoronoiDiagram() const
      {
          return M_teammates_voronoi_diagram;
      }

    const
    rcsc::VoronoiDiagram & passVoronoiDiagram() const
      {
          return M_pass_voronoi_diagram;
      }

    const
    PlayerGraph & fieldPlayerGraph() const
      {
          return M_field_player_graph;
      }

    void update( const rcsc::WorldModel & wm );


private:

    void updateVoronoiDiagram( const rcsc::WorldModel & wm );
    void updatePlayerGraph( const rcsc::WorldModel & wm );

    void writeDebugLog();

};

#endif
