// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_offensive_half.h"

#include "strategy.h"

#include "bhv_cross.h"
#include "bhv_basic_defensive_kick.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_keep_shoot_chance.h"

#include "bhv_basic_tackle.h"

#include "bhv_basic_move.h"
#include "bhv_attacker_offensive_move.h"
#include "bhv_block_dribble.h"

#include "body_advance_ball_test.h"
#include "bhv_pass_test.h"

#include <rcsc/action/body_clear_ball.h>
#include <rcsc/action/body_dribble2008.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_intercept.h>

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_scan_field.h>

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/player_config.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/ray_2d.h>
#include <rcsc/soccer_math.h>

#include "bhv_offensive_half_defensive_move.h"
#include "bhv_offensive_half_offensive_move.h"
#include "bhv_block_ball_owner.h"

#include "neck_offensive_intercept_neck.h"

using namespace rcsc;

const std::string RoleOffensiveHalf::NAME( "OffensiveHalf" );

/*-------------------------------------------------------------------*/
/*!

 */
namespace {
rcss::RegHolder role = SoccerRole::creators().autoReg( &RoleOffensiveHalf::create,
                                                       RoleOffensiveHalf::NAME );
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
RoleOffensiveHalf::execute( PlayerAgent * agent )
{
    //////////////////////////////////////////////////////////////
    // play_on play
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        // kick
        doKick( agent );
    }
    else
    {
        // positioning
        doMove( agent );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doKick( PlayerAgent * agent )
{
    const double nearest_opp_dist
        = agent->world().getDistOpponentNearestToSelf( 5 );

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
        if ( nearest_opp_dist < ServerParam::i().tackleDist() )
        {
            Body_AdvanceBallTest().execute( agent );
            agent->setNeckAction( new Neck_ScanField() );
        }
        else
        {
            Bhv_BasicOffensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_Danger:
        agent->debugClient().addMessage( "DangerArea" );
        if ( nearest_opp_dist < ServerParam::i().tackleDist() )
        {
            Body_ClearBall().execute( agent );
            agent->setNeckAction( new Neck_ScanField() );
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
        agent->debugClient().addMessage( "DribAtt" );
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_OffMidField:
        agent->debugClient().addMessage( "OffMid" );
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_Cross:
        agent->debugClient().addMessage( "CrossArea" );
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_ShootChance:
        doShootChanceKick( agent );
        break;
    default:
        dlog.addText( Logger::ROLE,
                            __FILE__": unknown ball area" );
        Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doMove( PlayerAgent * agent )
{
    switch ( Strategy::get_ball_area( agent->world() ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_OffensiveHalfDefensiveMove().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
        Bhv_OffensiveHalfOffensiveMove().execute( agent );
        break;
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
        doShootAreaMove( agent );
        break;
    default:
        dlog.addText( Logger::ROLE,
                            __FILE__": unknown ball area" );
        Bhv_BasicMove().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doShootChanceKick( PlayerAgent * agent )
{
    agent->debugClient().addMessage( "ShootChance" );
    dlog.addText( Logger::ROLE,
                        __FILE__": doShootChanceKick()" );

    const Bhv_PassTest::PassRoute * pass
        = Bhv_PassTest::get_best_pass_first( agent->world() );
    if ( pass )
    {
        dlog.addText( Logger::TEAM,
                            __FILE__":  pass found (%.1f, %.1f)",
                            pass->receive_point_.x,
                            pass->receive_point_.y );
        const PlayerObject * nearest_opp
            = agent->world().getOpponentNearestToSelf( 1 );

        if ( pass->receive_point_.x > 36.0
             && nearest_opp
             && nearest_opp->distFromSelf() < 2.0 )
        {
            agent->debugClient().addMessage( "KeepChancePass1" );
            Bhv_PassTest().execute( agent );
            return;
        }
    }

    Bhv_KeepShootChance().execute( agent );
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doShootAreaMove( PlayerAgent * agent )
{
    agent->debugClient().addMessage( "ShotAreaMove" );

    //----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.8, 90.0 ).execute( agent ) )
    {
        return;
    }

    const WorldModel & wm = agent->world();
    const Vector2D home_pos = Strategy::i().getPosition( wm.self().unum() );

    dlog.addText( Logger::ROLE,
                        __FILE__": doShootAreaMove()" );

    if ( doShootAreaIntercept( agent ) )
    {
        return;
    }

    if ( home_pos.x > 30.0
         && home_pos.absY() < 7.0 )
    {
        agent->debugClient().addMessage( "ShotSupport" );
        Bhv_AttackerOffensiveMove( false ).execute( agent );
        return;
    }

    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( self_min < mate_min )
    {
        agent->debugClient().addMessage( "GetBall" );
        Body_Intercept().execute( agent );
        agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
        return;
    }

    if ( wm.existKickableOpponent()
         && ( wm.teammatesFromBall().empty()
              || wm.teammatesFromBall().front()->distFromBall() > 2.0 )
         )
    {
        dlog.addText( Logger::ROLE,
                            __FILE__": doShootAreaMove() ball owner is opponent. change to normal move" );
        Bhv_BasicMove().execute( agent );
        return;
    }

    // set dash power
    static bool S_recover_mode = false;
    double dash_power;
    if ( wm.self().stamina()
         < ServerParam::i().recoverDecThrValue() + 300.0 )
    {
        S_recover_mode = true;
    }
    else if ( wm.self().stamina()
              > ServerParam::i().staminaMax() * 0.7 )
    {
        S_recover_mode = false;
    }

    // recover
    if ( S_recover_mode )
    {
        dlog.addText( Logger::ROLE,
                            __FILE__": doShootAreaMove() recovering" );
        dash_power
            = wm.self().playerType().staminaIncMax() * wm.self().recovery()
            - 25.0; // preffered recover value
        if ( dash_power < 0.0 ) dash_power = 0.0;
    }
    else if ( opp_min < self_min - 2
              && opp_min < mate_min - 5 )
    {
        dash_power = std::max( ServerParam::i().maxDashPower() * 0.2,
                               wm.self().playerType().staminaIncMax() * 0.8 );
    }
    else
    {
        dash_power = wm.self().getSafetyDashPower( ServerParam::i().maxDashPower() );
    }

    Vector2D target_point = home_pos;
    if ( wm.self().pos().dist( home_pos ) < 8.0 )
    {
        const PlayerObject * nearest_opp = wm.getOpponentNearestToSelf( 1 );
        // check opponent
        if ( nearest_opp
             && nearest_opp->pos().dist( home_pos ) < 3.0 )
        {
            if ( home_pos.y > 0.0 ) target_point.y -= 5.0;
            if ( home_pos.y < 0.0 ) target_point.y += 5.0;
        }
    }

    // move
    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.5 ) dist_thr = 0.5;

    agent->debugClient().addMessage( "ShootMove%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    agent->debugClient().addCircle( target_point, dist_thr );
    dlog.addText( Logger::ROLE,
                        __FILE__": doShootAreaMove() chance area move to (%.1f, %.1f) power=%.1f",
                        target_point.x, target_point.y, dash_power );

    if ( ! Body_GoToPoint( target_point, dist_thr, dash_power ).execute( agent ) )
    {
        // face to front or side
        AngleDeg body_angle = 0.0;
        if ( wm.ball().angleFromSelf().abs() > 100.0 )
        {
            body_angle = ( wm.ball().pos().y > wm.self().pos().y
                           ? 90.0
                           : -90.0 );
        }
        Body_TurnToAngle( body_angle ).execute( agent );
    }

    if ( wm.ball().distFromSelf() < ServerParam::i().visibleDistance() )
    {
        agent->setNeckAction( new Neck_TurnToLowConfTeammate() );
    }
    else
    {
        agent->setNeckAction( new Neck_TurnToBallOrScan() );
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
RoleOffensiveHalf::doShootAreaIntercept( PlayerAgent * agent )
{
    const WorldModel & wm = agent->world();
    const int self_min = wm.interceptTable()->selfReachCycle();
    const int mate_min = wm.interceptTable()->teammateReachCycle();
    const int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent()
         && self_min < mate_min
         && self_min < opp_min )
    {
        if ( Body_Intercept().execute( agent ) )
        {
            agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
            return true;
        }
    }

    return false;
}
