// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 3, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "role_side_back.h"

#include "strategy.h"

#include "bhv_cross.h"
#include "bhv_basic_defensive_kick.h"
#include "bhv_basic_offensive_kick.h"
#include "bhv_keep_shoot_chance.h"

#include "bhv_basic_tackle.h"

#include "bhv_basic_move.h"
#include "bhv_defender_basic_block_move.h"
#include "bhv_danger_area_tackle.h"
#include "bhv_block_dribble.h"

#include "body_advance_ball_test.h"
#include <rcsc/action/body_dribble2008.h>
#include <rcsc/action/body_intercept.h>
#include <rcsc/action/body_hold_ball.h>
#include "bhv_pass_test.h"

#include <rcsc/action/basic_actions.h>
#include <rcsc/action/bhv_go_to_point_look_ball.h>
#include <rcsc/action/bhv_scan_field.h>
#include <rcsc/action/body_go_to_point.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_player_or_scan.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/line_2d.h>
#include <rcsc/geom/ray_2d.h>
#include <rcsc/soccer_math.h>
#include <rcsc/math_util.h>

#include "bhv_side_back_cross_block_move.h"
#include "bhv_side_back_danger_move.h"
#include "bhv_side_back_defensive_move.h"
#include "bhv_side_back_stopper_move.h"
#include "bhv_mark_pass_line.h"

#include "neck_default_intercept_neck.h"
#include "neck_offensive_intercept_neck.h"

using namespace rcsc;

const std::string RoleSideBack::NAME( "SideBack" );

/*-------------------------------------------------------------------*/
/*!

 */
namespace {
rcss::RegHolder role = SoccerRole::creators().autoReg( &RoleSideBack::create,
                                                       RoleSideBack::NAME );
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
RoleSideBack::execute( PlayerAgent * agent )
{
    //////////////////////////////////////////////////////////////
    // play_on play
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        // kick
        doKick( agent );
    }
    else
    {
        // positioning
        doMove( agent );
    }

    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleSideBack::doKick( PlayerAgent* agent )
{
    const double nearest_opp_dist
        = agent->world().getDistOpponentNearestToSelf( 5 );

    switch ( Strategy::get_ball_area(agent->world().ball().pos()) ) {
    case Strategy::BA_CrossBlock:
        if ( nearest_opp_dist < 3.0 )
        {
            dlog.addText( Logger::ROLE,
                          __FILE__": doKick() danger" );
            if ( ! Bhv_PassTest().execute( agent ) )
            {
                Body_AdvanceBallTest().execute( agent );

                if ( agent->effector().queuedNextBallKickable() )
                {
                    agent->setNeckAction( new Neck_ScanField() );
                }
                else
                {
                    agent->setNeckAction( new Neck_TurnToBall() );
                }
            }
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
        if ( nearest_opp_dist < 3.0 )
        {
            dlog.addText( Logger::ROLE,
                          __FILE__": doKick(). very danger" );
            if ( ! Bhv_PassTest().execute( agent ) )
            {
                Body_AdvanceBallTest().execute( agent );

                if ( agent->effector().queuedNextBallKickable() )
                {
                    agent->setNeckAction( new Neck_ScanField() );
                }
                else
                {
                    agent->setNeckAction( new Neck_TurnToBall() );
                }
            }
        }
        else
        {
            Bhv_BasicDefensiveKick().execute( agent );
        }
        break;
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_BasicDefensiveKick().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_Cross:
        Bhv_Cross().execute( agent );
        break;
    case Strategy::BA_ShootChance:
        Bhv_KeepShootChance().execute( agent );
        break;
    default:
        dlog.addText( Logger::ROLE,
                      __FILE__": unknown ball area" );
        Body_HoldBall().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleSideBack::doMove( PlayerAgent * agent )
{
    switch ( Strategy::get_ball_area( agent->world() ) ) {
    case Strategy::BA_CrossBlock:
        Bhv_SideBackCrossBlockMove().execute( agent );
        break;
    case Strategy::BA_Stopper:
        Bhv_SideBackStopperMove().execute( agent );
        break;
    case Strategy::BA_Danger:
        Bhv_SideBackDangerMove().execute( agent );
        break;
    case Strategy::BA_DribbleBlock:
        Bhv_SideBackDefensiveMove().execute( agent );
        break;
    case Strategy::BA_DefMidField:
        Bhv_SideBackDefensiveMove().execute( agent );
        break;
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
        doBasicMove( agent );
        break;
    default:
        dlog.addText( Logger::ROLE,
                      __FILE__": unknown ball area" );
        Bhv_BasicMove().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleSideBack::doBasicMove( PlayerAgent * agent )
{
    dlog.addText( Logger::ROLE,
                  __FILE__": doBasicMove" );

    const WorldModel & wm = agent->world();
    const Vector2D home_pos = Strategy::i().getPosition( wm.self().unum() );


    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.85, 60.0 ).execute( agent ) )
    {
        return;
    }

    //--------------------------------------------------
    // intercept

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( self_min <= 3
         || ( self_min < 20
              && ( self_min < mate_min
                   || ( self_min < mate_min + 3 && mate_min > 3 ) )
              && self_min <= opp_min + 1 )
         || ( self_min < mate_min
              && opp_min >= 2
              && self_min <= opp_min + 1 )
         )
    {
        dlog.addText( Logger::ROLE,
                      __FILE__": doBasicMove() intercept" );
        Body_Intercept().execute( agent );

        if ( opp_min >= self_min + 3 )
        {
            agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
        }
        else
        {
            agent->setNeckAction( new Neck_DefaultInterceptNeck
                                  ( new Neck_TurnToBallOrScan() ) );
        }

        return;
    }

    /*--------------------------------------------------------*/

    if ( Bhv_MarkPassLine().execute( agent ) )
    {
        return;
    }

    // decide move target point
    Vector2D target_point = getBasicMoveTarget( agent, home_pos );

    // decide dash power
    double dash_power = Strategy::get_defender_dash_power( wm, target_point );
    // decide threshold
    double dist_thr = std::fabs( wm.ball().pos().x - wm.self().pos().x ) * 0.1; //wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 1.0 ) dist_thr = 1.0;

    agent->debugClient().addMessage( "%.0f", dash_power );
    agent->debugClient().setTarget( target_point );
    agent->debugClient().addCircle( target_point, dist_thr );

    if ( ! Body_GoToPoint( target_point, dist_thr, dash_power ).execute( agent ) )
    {
        AngleDeg body_angle = ( wm.ball().pos().y < wm.self().pos().y
                                ? -90.0
                                : 90.0 );
        Body_TurnToAngle( body_angle ).execute( agent );
    }

    if ( wm.ball().distFromSelf() < 20.0
         && ( wm.existKickableOpponent()
              || opp_min <= 3 )
         )
    {
        agent->setNeckAction( new Neck_TurnToBall() );
    }
    else
    {
        agent->setNeckAction( new Neck_TurnToBallOrScan() );
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
Vector2D
RoleSideBack::getBasicMoveTarget( PlayerAgent * agent,
                                  const Vector2D & home_pos )
{
    const WorldModel & wm = agent->world();

    Vector2D target_point = home_pos;

    const int opp_min = wm.interceptTable()->opponentReachCycle();

    const Vector2D opp_trap_pos = ( opp_min < 100
                                    ? wm.ball().inertiaPoint( opp_min )
                                    : wm.ball().inertiaPoint( 20 ) );

    // decide wait position
    // adjust to the defence line

    if ( wm.ball().pos().x < 10.0
         && -36.0 < home_pos.x
         && home_pos.x < 10.0
         && wm.self().pos().x > home_pos.x
         && wm.ball().pos().x > wm.self().pos().x + 3.0 )
    {
        // make static line
        double tmpx = home_pos.x;
        for ( double x = -12.0; x > -27.0; x -= 8.0 )
        {
            if ( opp_trap_pos.x > x )
            {
                tmpx = x - 3.3;
                break;
            }
        }

        if ( std::fabs( wm.self().pos().y - opp_trap_pos.y ) > 5.0 )
        {
            tmpx -= 3.0;
        }
        target_point.x = tmpx;

        agent->debugClient().addMessage( "LineDef%.1f", tmpx );
    }
    else
    {
        agent->debugClient().addMessage( "SB:Normal" );
    }

#if 0
    const bool is_left_side = Strategy::i().getPositionType( wm.self().unum() ) == Position_Left;
    const PlayerPtrCont::const_iterator end = wm.opponentsFromSelf().end();

    double first_x = 100.0;

    for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
          it != end;
          ++it )
    {
        if ( (*it)->pos().x < first_x ) first_x = (*it)->pos().x;
    }

    double max_y = 1.0;

    for ( PlayerPtrCont::const_iterator it = wm.opponentsFromSelf().begin();
          it != end;
          ++it )
    {
        if ( is_left_side && (*it)->pos().y > 0.0 ) continue;
        if ( ! is_left_side && (*it)->pos().y < 0.0 ) continue;

        //if ( (*it)->pos().x > target_point.x + 15.0 ) continue;
        if ( (*it)->pos().x > first_x + 15.0 ) continue;

        if ( (*it)->pos().absY() > max_y )
        {
            dlog.addText( Logger::ROLE,
                          __FILE__": getBasicMoveTarget. updated max y. opp=%d(%.1f %.1f)",
                          (*it)->unum(),
                          (*it)->pos().x, (*it)->pos().y );
            max_y = (*it)->pos().absY();
        }
    }

    if ( max_y > 0.0
         && max_y < target_point.absY() )
    {
        target_point.y = ( is_left_side
                           ? -max_y + 0.5
                           : max_y - 0.5 );
        dlog.addText( Logger::ROLE,
                      __FILE__": getBasicMoveTarget. shrink formation y. %.1f",
                      target_point.y );
        agent->debugClient().addMessage( "ShrinkY" );
    }
#endif

    return target_point;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleSideBack::doCrossBlockAreaMove( PlayerAgent * agent,
                                    const Vector2D & home_pos )
{
    dlog.addText( Logger::ROLE,
                  __FILE__": doCrossBlockAreaMove()" );

    //-----------------------------------------------
    // tackle
    if ( Bhv_DangerAreaTackle( 0.8 ).execute( agent ) )
    {
        return;
    }

    const WorldModel & wm = agent->world();

    //--------------------------------------------------
    // intercept
    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( ( self_min < mate_min && self_min < opp_min )
         || ( self_min <= 2 && wm.ball().pos().absY() < 20.0 ) )
    {
        dlog.addText( Logger::ROLE,
                      __FILE__": doCrossBlockAreaMove() intercept" );
        Body_Intercept().execute( agent );

        if ( opp_min >= self_min + 3 )
        {
            agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
        }
        else
        {
            agent->setNeckAction( new Neck_DefaultInterceptNeck
                                  ( new Neck_TurnToBallOrScan() ) );
        }

        return;
    }

    //-----------------------------------------
    // set move point

    Vector2D center( -41.5, 0.0 );

    int min_cyc = std::min( mate_min, opp_min );

    Vector2D ball_future
        = inertia_n_step_point( wm.ball().pos(),
                                wm.ball().vel(),
                                min_cyc,
                                ServerParam::i().ballDecay() );

    if ( ball_future.y * home_pos.y < 0.0 )
    {
        dlog.addText( Logger::ROLE,
                      __FILE__": doCrossBlockAreaMove() opposite side. basic move" );
        doBasicMove( agent );
        return;
    }

    if ( ball_future.x < -45.0 )
    {
        center.x = ball_future.x + 1.0;
        center.x = std::max( -48.0, center.x );
    }

    Line2D block_line( ball_future, center );

    Vector2D block_point = home_pos;
#if 0
#if 1
    block_point.y = ball_future.absY() - 8.0;
    block_point.y = std::max( block_point.y, 16.5 );
#elif 0
    block_point.y = ball_future.absY() - 7.0;
    block_point.y = std::max( block_point.y, 18.0 );
    block_point.y = std::min( block_point.y, home_pos.absY() );
    block_point.y = std::min( ball_future.absY() - 2.0, block_point.y );
#elif 0
    block_point.y = ball_future.absY() - 5.0;
    block_point.y = std::max( block_point.y, 20.0 );
    block_point.y = std::min( ball_future.absY() - 2.0, block_point.y );
#else
    block_point.y = ball_future.absY() - 10.0;
    block_point.y = std::max( block_point.y, 15.0 );
#endif
    if ( ball_future.y < 0.0 ) block_point.y *= -1.0;

    block_point.x = block_line.getX( block_point.y );

#if 1
    if ( block_point.x < wm.self().pos().x - 3.0 )
    {
        Ray2D body_ray( wm.self().pos(), wm.self().body() );
        Vector2D intersect = body_ray.intersection( block_line );
        if ( intersect.valid()
             && 18.0 < intersect.absY()
             && intersect.absY() < ball_future.absY() )
        {
            dlog.addText( Logger::ROLE,
                          __FILE__":  doCrossBlockAreaMove()"
                          " ball is back. changed to intersect. (%.1f %.1f)->(%.1f %.1f)",
                          block_point.x, block_point.y,
                          intersect.x, intersect.y );
            block_point = intersect;
        }
    }
#endif
#endif

    //------------------------------------------
    // set dash power
    double dash_power = ServerParam::i().maxDashPower();
    if ( wm.self().pos().x < block_point.x + 5.0 )
    {
        dash_power -= wm.ball().distFromSelf() * 2.0;
        dash_power = std::max( 20.0, dash_power );
    }

    double dist_thr = wm.ball().distFromSelf() * 0.07;
    if ( dist_thr < 0.8 ) dist_thr = 0.8;

    dlog.addText( Logger::ROLE,
                  __FILE__":  doCrossBlockAreaMove() go to (%.1f, %.1f) dash_powe=%.1f",
                  block_point.x, block_point.y, dash_power );
    agent->debugClient().addMessage( "CrossBlock%.0f", dash_power );
    agent->debugClient().setTarget( block_point );
    agent->debugClient().addCircle( block_point, dist_thr );

    if ( ! Body_GoToPoint( block_point,
                           dist_thr,
                           dash_power
                           ).execute( agent )
         )
    {
        AngleDeg body_angle = ( wm.ball().angleFromSelf().abs() < 80.0
                                ? 0.0
                                : 180.0 );
        Body_TurnToAngle( body_angle ).execute( agent );
    }

    agent->setNeckAction( new Neck_TurnToBall() );
}

/*-------------------------------------------------------------------*/
/*!

 */
void
RoleSideBack::doBlockSideAttacker( PlayerAgent * agent,
                                   const Vector2D & home_pos )
{
    dlog.addText( Logger::ROLE,
                  __FILE__": doBlockSideAttacker()" );

    const WorldModel & wm = agent->world();

    //-----------------------------------------------
    // tackle
    if ( Bhv_BasicTackle( 0.85, 60.0 ).execute( agent ) )
    {
        return;
    }

    //--------------------------------------------------
    // intercept

    int self_min = wm.interceptTable()->selfReachCycle();
    int mate_min = wm.interceptTable()->teammateReachCycle();
    int opp_min = wm.interceptTable()->opponentReachCycle();

    if ( self_min < opp_min && self_min < mate_min )
    {
        agent->debugClient().addMessage( "BlockGetBall(1)" );
        Body_Intercept().execute( agent );

        if ( opp_min >= self_min + 3 )
        {
            agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
        }
        else
        {
            agent->setNeckAction( new Neck_DefaultInterceptNeck
                                  ( new Neck_TurnToBallOrScan() ) );
        }

        return;
    }

    Vector2D ball_get_pos = wm.ball().inertiaPoint( self_min );
    Vector2D opp_reach_pos = wm.ball().inertiaPoint( opp_min );

    if ( ( ( self_min <= 4 && opp_min >= 4 )
           || self_min <= 2 )
         && wm.self().pos().x < ball_get_pos.x
         && wm.self().pos().x < wm.ball().pos().x - 0.2
         && ( std::fabs( wm.ball().pos().y - wm.self().pos().y ) < 1.5
              || wm.ball().pos().absY() < wm.self().pos().absY()
              )
         )
    {
        agent->debugClient().addMessage( "BlockGetBall(2)" );
        Body_Intercept().execute( agent );

        if ( opp_min >= self_min + 3 )
        {
            agent->setNeckAction( new Neck_OffensiveInterceptNeck() );
        }
        else
        {
            agent->setNeckAction( new Neck_DefaultInterceptNeck
                                  ( new Neck_TurnToBallOrScan() ) );
        }

        return;
    }

    if ( wm.self().pos().x < wm.ball().pos().x - 5.0
         || ( home_pos.x < wm.ball().pos().x - 5.0
              && wm.self().pos().x < home_pos.x )
         )
    {
        agent->debugClient().addMessage( "BlockToBasic" );
        doBasicMove( agent );
        return;
    }

    //--------------------------------------------------
    // block move
    if ( opp_min < mate_min - 1
         && opp_reach_pos.absY() > 20.0
         && opp_reach_pos.y * home_pos.y > 0.0 )
    {
        if ( Bhv_BlockDribble().execute( agent ) )
        {
            return;
        }
    }

    Vector2D block_point = home_pos;

    if ( wm.ball().pos().x < wm.self().pos().x )
    {
        agent->debugClient().addMessage( "SideBlockStatic" );

        block_point.x = -42.0;
        block_point.y = wm.self().pos().y;
        if ( wm.ball().pos().absY() < wm.self().pos().absY() - 0.5 )
        {
            agent->debugClient().addMessage( "Correct" );
            block_point.y = 10.0;
            if ( home_pos.y < 0.0 ) block_point.y *= -1.0;
        }
        dlog.addText( Logger::ROLE,
                      __FILE__": doBlockSideAttacker() static move" );
    }
    else
    {
        if ( wm.ball().pos().x > -30.0
             && wm.self().pos().x > home_pos.x
             && wm.ball().pos().x > wm.self().pos().x + 3.0 )
        {
            // make static line
            // 0, -10, -20, ...
            //double tmpx = std::floor( wm.ball().pos().x * 0.1 ) * 10.0 - 3.5; //3.0;
            double tmpx;
            if ( wm.ball().pos().x > -10.0 ) tmpx = -13.5;
            else if ( wm.ball().pos().x > -16.0 ) tmpx = -19.5;
            else if ( wm.ball().pos().x > -24.0 ) tmpx = -27.5;
            else if ( wm.ball().pos().x > -30.0 ) tmpx = -33.5;
            else tmpx = home_pos.x;

            if ( tmpx > 0.0 ) tmpx = 0.0;
            if ( tmpx > home_pos.x + 5.0 ) tmpx = home_pos.x + 5.0;

            if ( wm.defenseLineX() < tmpx - 1.0 )
            {
                tmpx = wm.defenseLineX() + 1.0;
            }

            block_point.x = tmpx;
        }
    }

    //-----------------------------------------
    // set dash power
    double dash_power = ServerParam::i().maxDashPower();
    if ( wm.self().pos().x < wm.ball().pos().x - 8.0 )
    {
        dash_power = wm.self().playerType().staminaIncMax() * 0.8;
    }
    else if ( wm.self().pos().x < wm.ball().pos().x
              && wm.self().pos().absY() < wm.ball().pos().absY() - 1.0
              && wm.self().stamina() < ServerParam::i().staminaMax() * 0.65 )
    {
        dash_power = ServerParam::i().maxDashPower() * 0.7;
    }

    agent->debugClient().addMessage( "BlockSideAttack%.0f", dash_power );

    double dist_thr = wm.ball().distFromSelf() * 0.1;
    if ( dist_thr < 0.8 ) dist_thr = 0.8;

    agent->debugClient().setTarget( block_point );
    agent->debugClient().addCircle( block_point, dist_thr );

    if ( ! Body_GoToPoint( block_point, dist_thr, dash_power ).execute( agent ) )
    {
        Vector2D face_point( -48.0, 0.0 );
        Body_TurnToPoint( face_point ).execute( agent );
    }

    agent->setNeckAction( new Neck_TurnToBallOrScan() );
}
